%if 0
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
%endif

/* Ctrl-Tab */

#ctrlTab-panel {
  -moz-appearance: none;
%ifdef XP_MACOSX
  -moz-window-shadow: none;
%endif
  background: hsla(0,0%,33%,.85);
  color: white;
  border-style: none;
  padding: 20px 10px 10px;
%ifndef XP_MACOSX
  font-weight: bold;
%endif
  text-shadow: 0 0 1px hsl(0,0%,12%), 0 0 2px hsl(0,0%,12%);
}

.ctrlTab-favicon[src] {
  background-color: white;
  width: 20px;
  height: 20px;
  padding: 2px;
}

.ctrlTab-preview-inner > .tabPreview-canvas {
  box-shadow: 1px 1px 2px hsl(0,0%,12%);
}

.ctrlTab-preview:not(#ctrlTab-showAll) > * > .ctrlTab-preview-inner > .tabPreview-canvas {
  margin-bottom: 2px;
}

.ctrlTab-preview-inner {
  padding: 8px;
  border: 2px solid transparent;
  border-radius: .5em;
}

.ctrlTab-preview:not(#ctrlTab-showAll) > * > .ctrlTab-preview-inner {
  margin: -10px -10px 0;
}

#ctrlTab-showAll:not(:focus) > * > .ctrlTab-preview-inner {
  background-color: rgba(255,255,255,.2);
}

.ctrlTab-preview:focus > * > .ctrlTab-preview-inner {
  color: white;
  background-color: rgba(0,0,0,.6);
  text-shadow: none;
  border-color: white;
}

#ctrlTab-showAll {
  margin-top: .5em;
}

