#context-navigation > .menuitem-iconic {
  -moz-box-flex: 1;
  -moz-box-pack: center;
  -moz-box-align: center;
}

#context-navigation > .menuitem-iconic > .menu-iconic-left {
  -moz-appearance: none;
}

#context-navigation > .menuitem-iconic > .menu-iconic-left > .menu-iconic-icon {
  width: 16px;
  height: 16px;
  margin: 7px;
  filter: url(chrome://browser/skin/filters.svg#fill);
  fill: currentColor;
}

#context-back {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#back");
}

#context-forward {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#forward");
}

#context-reload {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#reload");
}

#context-stop {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#stop");
}

#context-bookmarkpage {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#bookmark");
}

#context-bookmarkpage[starred=true] {
  list-style-image: url("chrome://browser/skin/content-contextmenu.svg#bookmarked");
}

#context-back:-moz-locale-dir(rtl),
#context-forward:-moz-locale-dir(rtl),
#context-reload:-moz-locale-dir(rtl) {
  transform: scaleX(-1);
}

#context-media-eme-learnmore {
  list-style-image: url("chrome://browser/skin/drm-icon.svg#chains");
}
