Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/AddonManager.jsm");

function install(data, reason) {}

function startup(data, reason) {}

function shutdown(data, reason) {}
